
class TorrentoolException(Exception):
    """Base torrentool exception. All others are inherited from it."""


class BencodeError(TorrentoolException):
    """Base exception for bencode related errors."""


class BencodeDecodingError(BencodeError):
    """Raised when torrentool is unable to decode bencoded data."""


class BencodeEncodingError(BencodeError):
    """Raised when torrentool is unable to encode data into bencode."""


class TorrentError(TorrentoolException):
    """Base exception for Torrent object related errors."""


class RemoteUploadError(TorrentoolException):
    """Base class for upload to remotes related issues."""


class RemoteDownloadError(TorrentoolException):
    """Base class for issues related to downloads from remotes."""



"""
Exposes commonly used classes and functions.

"""
from .bencode import Bencode  # noqa
from .torrent import Torrent  # noqa
from .utils import upload_to_cache_server, get_open_trackers_from_local, get_open_trackers_from_remote  # noqa



from os import path, getcwd

import click

from . import VERSION
from .api import Torrent
from .exceptions import RemoteUploadError, RemoteDownloadError
from .utils import humanize_filesize, upload_to_cache_server, get_open_trackers_from_remote, \
    get_open_trackers_from_local


@click.group()
@click.version_option(version='.'.join(map(str, VERSION)))
def start():
    """Torrentool command line utilities."""


@start.group()
def torrent():
    """Torrent-related commands."""


@torrent.command()
@click.argument('torrent_path', type=click.Path(exists=True, writable=False, dir_okay=False))
def info(torrent_path):
    """Print out information from .torrent file."""

    my_torrent = Torrent.from_file(torrent_path)

    size = my_torrent.total_size

    click.secho(f'Name: {my_torrent.name}', fg='blue')
    click.secho('Files:')
    for file_tuple in my_torrent.files:
        click.secho(file_tuple.name)

    click.secho(f'Hash: {my_torrent.info_hash}', fg='blue')
    click.secho(f'Size: {humanize_filesize(size)} ({size})', fg='blue')
    click.secho(f'Magnet: {my_torrent.get_magnet()}', fg='yellow')


@torrent.command()
@click.argument('source', type=click.Path(exists=True, writable=False))
@click.option('--dest', default=getcwd, type=click.Path(file_okay=False), help='Destination path to put .torrent file into. Default: current directory.')
@click.option('--tracker', default=None, help='Tracker announce URL (multiple comma-separated values supported).')
@click.option('--open_trackers', default=False, is_flag=True, help='Add open trackers announce URLs.')
@click.option('--comment', default=None, help='Arbitrary comment.')
@click.option('--cache', default=False, is_flag=True, help='Upload file to torrent cache services.')
def create(source, dest, tracker, open_trackers, comment, cache):
    """Create torrent file from a single file or a directory."""

    source_title = path.basename(source).replace('.', '_').replace(' ', '_')
    dest = f'{path.join(dest, source_title)}.torrent'

    click.secho(f'Creating torrent from {source} ...')

    my_torrent = Torrent.create_from(source)

    if comment:
        my_torrent.comment = comment

    urls = []

    if tracker:
        urls = tracker.split(',')

    if open_trackers:
        click.secho('Fetching an up-to-date open tracker list ...')
        try:
            urls.extend(get_open_trackers_from_remote())
        except RemoteDownloadError:
            click.secho('Failed. Using built-in open tracker list.', fg='red', err=True)
            urls.extend(get_open_trackers_from_local())

    if urls:
        my_torrent.announce_urls = urls

    my_torrent.to_file(dest)

    click.secho(f'Torrent file created: {dest}', fg='green')
    click.secho(f'Torrent info hash: {my_torrent.info_hash}', fg='blue')

    if cache:
        click.secho('Uploading to cache service ...')
        try:
            result = upload_to_cache_server(dest)
            click.secho(f'Cached torrent URL: {result}', fg='yellow')

        except RemoteUploadError as e:
            click.secho(f'Failed: {e}', fg='red', err=True)


def main():
    start(obj={})



import math
from os import path
from typing import List

from .exceptions import RemoteUploadError, RemoteDownloadError


OPEN_TRACKERS_FILENAME = 'open_trackers.ini'
REMOTE_TIMEOUT = 4


def get_app_version() -> str:
    """Returns full version string including application name
    suitable for putting into Torrent.created_by.

    """
    from torrentool import VERSION_STR
    return f'torrentool/{VERSION_STR}'


def humanize_filesize(bytes_size: int) -> str:
    """Returns human readable filesize.

    :param bytes_size:

    """
    if not bytes_size:
        return '0 B'
    
    names = ('B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB')

    name_idx = int(math.floor(math.log(bytes_size, 1024)))
    size = round(bytes_size / math.pow(1024, name_idx), 2)

    return f'{size} {names[name_idx]}'


def upload_to_cache_server(fpath: str) -> str:
    """Uploads .torrent file to a cache server.
    Returns upload file URL.

    :param fpath: File to upload

    """
    url_base = 'http://torrage.info'
    url_upload = f'{url_base}/autoupload.php'
    url_download = f'{url_base}/torrent.php?h='
    file_field = 'torrent'

    try:
        import requests

        response = requests.post(url_upload, files={file_field: open(fpath, 'rb')}, timeout=REMOTE_TIMEOUT)
        response.raise_for_status()

        info_cache = response.text
        return url_download + info_cache

    except (ImportError, requests.RequestException) as e:

        # Now trace is lost. `raise from` to consider.
        raise RemoteUploadError(f'Unable to upload to {url_upload}: {e}')


def get_open_trackers_from_remote() -> List[str]:
    """Returns open trackers announce URLs list from remote repo."""

    url_base = 'https://raw.githubusercontent.com/idlesign/torrentool/master/torrentool/repo'
    url = f'{url_base}/{OPEN_TRACKERS_FILENAME}'

    try:
        import requests

        response = requests.get(url, timeout=REMOTE_TIMEOUT)
        response.raise_for_status()

        open_trackers = response.text.splitlines()

    except (ImportError, requests.RequestException) as e:

        # Now trace is lost. `raise from` to consider.
        raise RemoteDownloadError(f'Unable to download from {url}: {e}')

    return open_trackers


def get_open_trackers_from_local() -> List[str]:
    """Returns open trackers announce URLs list from local backup."""

    with open(path.join(path.dirname(__file__), 'repo', OPEN_TRACKERS_FILENAME)) as f:
        open_trackers = map(str.strip, f.readlines())

    return list(open_trackers)



from codecs import encode
from operator import itemgetter
from pathlib import Path
from typing import Union, Tuple, Set

from .exceptions import BencodeDecodingError, BencodeEncodingError

TypeEncodable = Union[str, int, list, set, tuple, dict, bytes, bytearray]


class Bencode:
    """Exposes utilities for bencoding."""

    @classmethod
    def encode(cls, value: TypeEncodable) -> bytes:
        """Encodes a value into bencoded bytes.

        :param value: Python object to be encoded (str, int, list, dict).

        """
        val_encoding = 'utf-8'

        def encode_str(v: str) -> bytes:
            v_enc = encode(v, val_encoding)
            prefix = encode(f'{len(v_enc)}:', val_encoding)
            return prefix + v_enc

        def encode_(val: TypeEncodable) -> bytes:
            if isinstance(val, str):
                result = encode_str(val)

            elif isinstance(val, int):
                result = encode(f'i{val}e', val_encoding)

            elif isinstance(val, (list, set, tuple)):
                result = encode('l', val_encoding)
                for item in val:
                    result += encode_(item)
                result += encode('e', val_encoding)

            elif isinstance(val, dict):
                result = encode('d', val_encoding)

                # Dictionaries are expected to be sorted by key.
                for k, v in sorted(val.items(), key=itemgetter(0)):
                    result += (encode_str(k) + encode_(v))

                result += encode('e', val_encoding)

            elif isinstance(val, (bytes, bytearray)):
                result = encode(f'{len(val)}:', val_encoding)
                result += val

            else:
                raise BencodeEncodingError(f'Unable to encode `{type(val)}` {val}')

            return result

        return encode_(value)

    @classmethod
    def decode(cls, encoded: bytes, *, byte_keys: Set[str] = None) -> TypeEncodable:
        """Decodes bencoded data introduced as bytes.

        Returns decoded structure(s).

        :param encoded:

        :param byte_keys: Keys values for which should be treated
            as bytes (as opposed to UTF-8 strings).

        """
        def create_dict(items) -> dict:
            # Let's guarantee that dictionaries are sorted.
            k_v_pair = zip(*[iter(items)] * 2)
            return dict(sorted(k_v_pair, key=itemgetter(0)))

        def create_list(items) -> list:
            return list(items)

        stack_items = []
        stack_containers = []

        def compress_stack():
            target_container = stack_containers.pop()
            subitems = []

            while True:
                subitem = stack_items.pop()
                subitems.append(subitem)
                if subitem is target_container:
                    break

            container_creator = subitems.pop()
            container = container_creator(reversed(subitems))
            stack_items.append(container)

        def parse_forward(till_char: str, sequence: bytes) -> Tuple[int, int]:
            number = ''
            char_sub_idx = 0

            for char_sub_idx, char_sub in enumerate(sequence):
                char_sub = chr(char_sub)
                if char_sub == till_char:
                    break

                number += char_sub

            number = int(number or 0)
            char_sub_idx += 1

            return number, char_sub_idx

        while encoded:
            char = encoded[0]
            char = chr(char)

            if char == 'd':  # Dictionary
                stack_items.append(create_dict)
                stack_containers.append(create_dict)
                encoded = encoded[1:]

            elif char == 'l':  # List
                stack_items.append(create_list)
                stack_containers.append(create_list)
                encoded = encoded[1:]

            elif char == 'i':  # Integer
                number, char_sub_idx = parse_forward('e', encoded[1:])
                char_sub_idx += 1

                stack_items.append(number)
                encoded = encoded[char_sub_idx:]

            elif char.isdigit():  # String
                str_len, char_sub_idx = parse_forward(':', encoded)
                last_char_idx = char_sub_idx + str_len

                string = encoded[char_sub_idx:last_char_idx]
                try:
                    string = string.decode()

                except UnicodeDecodeError:

                    if stack_items:
                        latest_stack_item = stack_items[-1]
                    else:
                        latest_stack_item = None

                    if byte_keys is None or str(latest_stack_item) in byte_keys:
                        # Considered to be a bytestring (e.g. `pieces` hashes concatenation).
                        pass

                    else:
                        # Try to decode from UTF-8 with fallback to replace
                        # for non-standard .torrent files.
                        string = string.decode(errors='replace')

                stack_items.append(string)
                encoded = encoded[last_char_idx:]

            elif char == 'e':  # End of a dictionary or a list.
                try:
                    compress_stack()

                except IndexError as e:
                    raise BencodeDecodingError(
                        f'Unable to parse the rest of the data: "{encoded.decode(errors="replace")[:60]}"'
                    ) from e

                encoded = encoded[1:]

            else:
                raise BencodeDecodingError(f'Unable to interpret `{char}` char.')

        if len(stack_items) == 1:
            stack_items = stack_items.pop()

        return stack_items

    @classmethod
    def read_string(cls, string: Union[str, bytes], *, byte_keys: Set[str] = None) -> TypeEncodable:
        """Decodes a given bencoded string or bytestring.

        Returns decoded structure(s).

        :param string:

        :param byte_keys: Keys values for which should be treated
            as bytes (as opposed to UTF-8 strings).

        """
        if not isinstance(string, (bytes, bytearray)):
            string = string.encode()

        return cls.decode(string, byte_keys=byte_keys)

    @classmethod
    def read_file(cls, filepath: Union[str, Path], *, byte_keys: Set[str] = None) -> TypeEncodable:
        """Decodes bencoded data of a given file.

        Returns decoded structure(s).

        :param filepath:

        :param byte_keys: Keys values for which should be treated
            as bytes (as opposed to UTF-8 strings).

        """
        with open(str(filepath), mode='rb') as f:
            contents = f.read()

        return cls.decode(contents, byte_keys=byte_keys)



from calendar import timegm
from datetime import datetime
from functools import reduce
from hashlib import sha1
from os import walk, sep
from os.path import join, getsize, normpath
from pathlib import Path
from typing import List, Union, Optional, Tuple, NamedTuple
from urllib.parse import urlencode

from .bencode import Bencode
from .exceptions import TorrentError
from .utils import get_app_version

_ITERABLE_TYPES = (list, tuple, set)


class TorrentFile(NamedTuple):
    """Represents a file in torrent."""

    name: str
    length: int


class Torrent:
    """Represents a torrent file, and exposes utilities to work with it."""

    def __init__(self, dict_struct: dict = None):
        dict_struct: dict = dict_struct or {'info': {}}
        self._struct = dict_struct
        self._filepath: Optional[Path] = None

    def __str__(self):
        return f'Torrent: {self.name}'

    def _list_getter(self, key) -> list:
        return self._struct.get(key) or []

    def _list_setter(self, key, val):
        if val is None:
            try:
                del self._struct[key]
                return
            except KeyError:
                return

        if not isinstance(val, _ITERABLE_TYPES):
            val = [val]

        self._struct[key] = val

    @property
    def webseeds(self) -> List[str]:
        """A list of URLs where torrent data can be retrieved.

        See also: Torrent.httpseeds

        http://bittorrent.org/beps/bep_0019.html

        """
        return self._list_getter('url-list')

    @webseeds.setter
    def webseeds(self, val: List[str]):
        self._list_setter('url-list', val)

    @property
    def httpseeds(self) -> List[str]:
        """A list of URLs where torrent data can be retrieved.

        See also and prefer Torrent.webseeds

        http://bittorrent.org/beps/bep_0017.html

        """
        return self._list_getter('httpseeds')

    @httpseeds.setter
    def httpseeds(self, val: List[str]):
        self._list_setter('httpseeds', val)

    @property
    def files(self) -> List['TorrentFile']:
        """Files in torrent.

        List of namedtuples (filepath, size).

        """
        files = []
        info = self._struct.get('info')

        if not info:
            return files

        if 'files' in info:
            base = info['name']

            for f in info['files']:
                files.append(TorrentFile(join(base, *f['path']), f['length']))

        else:
            files.append(TorrentFile(info['name'], info['length']))

        return files

    @property
    def total_size(self) -> int:
        """Total size of all files in torrent."""
        return reduce(lambda prev, curr: prev + curr[1], self.files, 0)

    @property
    def info_hash(self) -> Optional[str]:
        """Hash of torrent file info section. Also known as torrent hash."""
        info = self._struct.get('info')

        if not info:
            return None

        return sha1(Bencode.encode(info)).hexdigest()

    @property
    def magnet_link(self) -> str:
        """Magnet link using BTIH (BitTorrent Info Hash) URN."""
        return self.get_magnet(detailed=False)

    @property
    def announce_urls(self) -> Optional[List[List[str]]]:
        """List of lists of announce (tracker) URLs.

        First inner list is considered as primary announcers list,
        the following lists as back-ups.

        http://bittorrent.org/beps/bep_0012.html

        """
        urls = self._struct.get('announce-list')

        if not urls:
            urls = self._struct.get('announce')
            if not urls:
                return []
            urls = [[urls]]

        return urls

    @announce_urls.setter
    def announce_urls(self, val: List[str]):
        self._struct['announce'] = ''
        self._struct['announce-list'] = []

        def set_single(val):
            del self._struct['announce-list']
            self._struct['announce'] = val

        if isinstance(val, _ITERABLE_TYPES):
            length = len(val)

            if length:
                if length == 1:
                    set_single(val[0])
                else:
                    for item in val:
                        if not isinstance(item, _ITERABLE_TYPES):
                            item = [item]
                        self._struct['announce-list'].append(item)
                    self._struct['announce'] = val[0]

        else:
            set_single(val)

    @property
    def comment(self) -> Optional[str]:
        """Optional. Free-form textual comments of the author."""
        return self._struct.get('comment')

    @comment.setter
    def comment(self, val: str):
        self._struct['comment'] = val
    
    @property
    def source(self) -> Optional[str]:
        """Optional. Often used by private trackers to create a unique infohash to prevent peer-leak."""
        return self._struct.get('info').get("source")

    @source.setter
    def source(self, val: str):
        self._struct['info']['source'] = val
    
    @property
    def creation_date(self) -> Optional[datetime]:
        """Optional. The creation time of the torrent, in standard UNIX epoch format. UTC."""

        date = self._struct.get('creation date')
        if date is not None:
            date = datetime.utcfromtimestamp(int(date))

        return date

    @creation_date.setter
    def creation_date(self, val: datetime):
        self._struct['creation date'] = timegm(val.timetuple())

    @property
    def created_by(self) -> Optional[str]:
        """Optional. Name and version of the program used to create the .torrent"""
        return self._struct.get('created by')

    @created_by.setter
    def created_by(self, val: str):
        self._struct['created by'] = val

    @property
    def private(self) -> bool:
        """Optional. If True the client MUST publish its presence to get other peers
        ONLY via the trackers explicitly described in the metainfo file. If False or is not present,
        the client may obtain peer from other means, e.g. PEX peer exchange, dht.

        """
        return self._struct.get('info', {}).get('private', False)

    @private.setter
    def private(self, val: bool):
        if not val:
            try:
                del self._struct['info']['private']
            except KeyError:
                pass
        else:
            self._struct['info']['private'] = 1

    @property
    def name(self) -> Optional[str]:
        """Torrent name (title)."""
        return self._struct.get('info', {}).get('name', None)

    @name.setter
    def name(self, val: str):
        self._struct['info']['name'] = val

    def get_magnet(self, detailed: Union[bool, list, tuple, set] = True) -> str:
        """Returns torrent magnet link, consisting of BTIH (BitTorrent Info Hash) URN
        anr optional other information.

        :param detailed:
            For boolean - whether additional info (such as trackers) should be included.
            For iterable - expected allowed parameter names:
                tr - trackers
                ws - webseeds

        """
        result = 'magnet:?xt=urn:btih:' + self.info_hash

        def add_tr():
            urls = self.announce_urls
            if not urls:
                return

            trackers = []

            urls = urls[0]  # Only primary announcers are enough.
            for url in urls:
                trackers.append(('tr', url))

            if trackers:
                return urlencode(trackers)

        def add_ws():
            webseeds = [('ws', url) for url in self.webseeds]
            if webseeds:
                return urlencode(webseeds)

        params_map = {
            'tr': add_tr,
            'ws': add_ws,
        }

        if detailed:
            details = []

            if isinstance(detailed, _ITERABLE_TYPES):
                requested_params = detailed
            else:
                requested_params = params_map.keys()

            for param in requested_params:
                param_val = params_map[param]()
                param_val and details.append(param_val)

            if details:
                result += f'&{"&".join(details)}'

        return result

    def to_file(self, filepath: str = None):
        """Writes Torrent object into file, either

        :param filepath:

        """
        if filepath is None and self._filepath is None:
            raise TorrentError('Unable to save torrent to file: no filepath supplied.')

        if filepath is not None:
            self._filepath = filepath

        with open(self._filepath, mode='wb') as f:
            f.write(self.to_string())

    def to_string(self) -> bytes:
        """Returns bytes representing torrent file."""
        return Bencode.encode(self._struct)

    @classmethod
    def _get_target_files_info(cls, src_path: Path) -> Tuple[List[Tuple[str, int, List[str]]], int]:
        is_dir = src_path.is_dir()

        src_path = f'{src_path}'  # Force walk() to return unicode names.
        target_files = []

        if is_dir:
            for base, _, files in walk(src_path):
                target_files.extend([join(base, fname) for fname in sorted(files)])

        else:
            target_files.append(src_path)

        target_files_ = []
        total_size = 0

        for fpath in target_files:
            file_size = getsize(fpath)

            if not file_size:
                continue

            target_files_.append((fpath, file_size, normpath(fpath.replace(src_path, '')).strip(sep).split(sep)))
            total_size += file_size

        return target_files_, total_size

    @classmethod
    def create_from(cls, src_path: Union[str, Path]) -> 'Torrent':
        """Returns Torrent object created from a file or a directory.

        :param src_path:

        """
        if isinstance(src_path, str):
            src_path = Path(src_path)

        target_files, size_data = cls._get_target_files_info(src_path)

        size_min = 32768  # 32 KiB
        size_default = 262144  # 256 KiB
        size_max = 1048576  # 1 MiB

        # todo use those limits as advised
        # chunks_min = 1000
        # chunks_max = 2200

        size_piece = size_min
        if size_data > size_min:
            size_piece = size_default

        if size_piece > size_max:
            size_piece = size_max

        def read(filepath):
            with open(filepath, 'rb') as f:
                while True:
                    chunk = f.read(size_piece - len(pieces_buffer))
                    chunk_size = len(chunk)
                    if chunk_size == 0:
                        break
                    yield chunk

        pieces = bytearray()
        pieces_buffer = bytearray()

        for fpath, _, _ in target_files:
            for chunk in read(fpath):
                pieces_buffer += chunk

                if len(pieces_buffer) == size_piece:
                    pieces += sha1(pieces_buffer).digest()[:20]
                    pieces_buffer = bytearray()

        if len(pieces_buffer):
            pieces += sha1(pieces_buffer).digest()[:20]
            pieces_buffer = bytearray()

        info = {
            'name': src_path.name,
            'pieces': bytes(pieces),
            'piece length': size_piece,
        }

        if src_path.is_dir():
            files = []

            for _, length, path in target_files:
                files.append({'length': length, 'path': path})

            info['files'] = files

        else:
            try:
                info['length'] = target_files[0][1]

            except IndexError:
                # Since empty files are skipped.
                raise TorrentError('Unable to create torrent for an empty file.')

        torrent = cls({'info': info})
        torrent.created_by = get_app_version()
        torrent.creation_date = datetime.utcnow()

        return torrent

    @classmethod
    def from_string(cls, string: str) -> 'Torrent':
        """Alternative constructor to get Torrent object from string.

        :param string:

        """
        return cls(Bencode.read_string(string, byte_keys={'pieces'}))

    @classmethod
    def from_file(cls, filepath: Union[str, Path]) -> 'Torrent':
        """Alternative constructor to get Torrent object from file.

        :param filepath:

        """
        if isinstance(filepath, str):
            filepath = Path(filepath)

        torrent = cls(Bencode.read_file(filepath, byte_keys={'pieces'}))
        torrent._filepath = filepath
        return torrent
